@@ -1,20 +1,30 @@
 import click
 import questionary as q
-import IPython
 import docker
 import os
 import time
+import subprocess
+from docker.client import DockerClient
 
 from threading import Thread
 from functools import wraps
-from traitlets.config import get_config
 from colorama import (Fore, Style)
 from sqlalchemy.engine.url import make_url
 
-from vantage6.common import (echo, info, warning, error,
+from vantage6.common import (info, warning, error, debug as debug_msg,
                              check_config_write_permissions)
-from vantage6.common.docker_addons import pull_if_newer
-from vantage6.common.globals import APPNAME, STRING_ENCODING
+from vantage6.common.docker_addons import (
+    pull_if_newer, remove_container_if_exists,
+    get_server_config_name
+)
+from vantage6.common.network_manager import NetworkManager
+from vantage6.common.globals import (
+    APPNAME,
+    STRING_ENCODING,
+    DEFAULT_DOCKER_REGISTRY,
+    DEFAULT_SERVER_IMAGE
+)
+
 # from vantage6.cli import fixture
 from vantage6.cli.globals import (DEFAULT_SERVER_ENVIRONMENT,
                                   DEFAULT_SERVER_SYSTEM_FOLDERS)
@@ -23,6 +33,7 @@ from vantage6.cli.configuration_wizard import (
     select_configuration_questionaire,
     configuration_wizard
 )
+from vantage6.cli.rabbitmq.queue_manager import RabbitMQManager
 from vantage6.cli import __version__
 
 
@@ -99,13 +110,18 @@ def cli_server():
 @cli_server.command(name='start')
 @click.option('--ip', default=None, help='ip address to listen on')
 @click.option('-p', '--port', default=None, type=int, help='port to listen on')
-@click.option('--debug', is_flag=True,
-              help='run server in debug mode (auto-restart)')
-@click.option('-i', '--image', default=None, help="Node Docker image to use")
+@click.option('-i', '--image', default=None, help="Server Docker image to use")
+@click.option('--rabbitmq-image', default=None,
+              help="RabbitMQ docker image to use")
 @click.option('--keep/--auto-remove', default=False,
               help="Keep image after finishing")
+@click.option('--mount-src', default='',
+              help="mount vantage6-master package source")
+@click.option('--attach/--detach', default=False,
+              help="Attach server logs to the console after start")
 @click_insert_context
-def cli_server_start(ctx, ip, port, debug, image, keep):
+def cli_server_start(ctx, ip, port, image, rabbitmq_image, keep, mount_src,
+                     attach):
     """Start the server."""
 
     info("Starting server...")
@@ -123,11 +139,13 @@ def cli_server_start(ctx, ip, port, debug, image, keep):
                   "is already running")
             exit(1)
 
-    # pull the server docker image
+    # Determine image-name. First we check if the option --image has been used.
+    # Then we check if the image has been specified in the config file, and
+    # finally we use the default settings from the package.
     if image is None:
         image = ctx.config.get(
             "image",
-            "harbor.vantage6.ai/infrastructure/server:latest"
+            f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_SERVER_IMAGE}"
         )
     info(f"Pulling latest server image '{image}'.")
     try:
@@ -139,17 +157,22 @@ def cli_server_start(ctx, ip, port, debug, image, keep):
         info(" ... success!")
 
     info("Creating mounts")
+    config_file = "/mnt/config.yaml"
     mounts = [
         docker.types.Mount(
-            "/mnt/config.yaml", str(ctx.config_file), type="bind"
+            config_file, str(ctx.config_file), type="bind"
         )
     ]
 
+    if mount_src:
+        mount_src = os.path.abspath(mount_src)
+        mounts.append(docker.types.Mount("/vantage6", mount_src, type="bind"))
+
     # FIXME: code duplication with cli_server_import()
     # try to mount database
     uri = ctx.config['uri']
     url = make_url(uri)
-    environment_vars = None
+    environment_vars = {"VANTAGE6_CONFIG_NAME": ctx.config_file_name}
 
     # If host is None, we're dealing with a file-based DB, like SQLite
     if (url.host is None):
@@ -168,19 +191,34 @@ def cli_server_start(ctx, ip, port, debug, image, keep):
             "/mnt/database/", dirname, type="bind"
         ))
 
-        environment_vars = {
-            "VANTAGE6_DB_URI": f"sqlite:////mnt/database/{basename}"
-        }
-
+        environment_vars["VANTAGE6_DB_URI"] = \
+            f"sqlite:////mnt/database/{basename}"
     else:
         warning(f"Database could not be transfered, make sure {url.host} "
                 "is reachable from the Docker container")
         info("Consider using the docker-compose method to start a server")
 
-    ip_ = f"--ip {ip}" if ip else ""
-    port_ = f"--port {port}" if port else ""
-    cmd = f'vserver-local start -c /mnt/config.yaml -e {ctx.environment} ' \
-          f'{ip_} {port_}'
+    # Create a docker network for the server and other services like RabbitMQ
+    # to reside in
+    server_network_mgr = NetworkManager(
+        network_name=f"{APPNAME}-{ctx.name}-{ctx.scope}-network"
+    )
+    server_network_mgr.create_network(is_internal=False)
+
+    # Note that ctx.data_dir has been created at this point, which is required
+    # for putting some RabbitMQ configuration files inside
+    info('Starting RabbitMQ container')
+    _start_rabbitmq(ctx, rabbitmq_image, server_network_mgr)
+
+    # The `ip` and `port` refer here to the ip and port within the container.
+    # So we do not really care that is it listening on all interfaces.
+    internal_port = 5000
+    cmd = (
+        f'uwsgi --http :{internal_port} --gevent 1000 --http-websockets '
+        '--master --callable app --disable-logging '
+        '--wsgi-file /vantage6/vantage6-server/vantage6/server/wsgi.py '
+        f'--pyargv {config_file}'
+    )
     info(cmd)
 
     info("Run Docker container")
@@ -195,14 +233,39 @@ def cli_server_start(ctx, ip, port, debug, image, keep):
             "name": ctx.config_file_name
         },
         environment=environment_vars,
-        ports={f"{port_}/tcp": ("127.0.0.1", port_)},
+        ports={f"{internal_port}/tcp": (ip, port_)},
         name=ctx.docker_container_name,
         auto_remove=not keep,
-        tty=True
+        tty=True,
+        network=server_network_mgr.network_name
     )
 
     info(f"Success! container id = {container}")
 
+    if attach:
+        logs = container.attach(stream=True, logs=True, stdout=True)
+        Thread(target=print_log_worker, args=(logs,), daemon=True).start()
+        while True:
+            try:
+                time.sleep(1)
+            except KeyboardInterrupt:
+                info("Closing log file. Keyboard Interrupt.")
+                exit(0)
+
+
+def _start_rabbitmq(ctx: ServerContext, rabbitmq_image: str,
+                    network_mgr: NetworkManager) -> None:
+    """ Starts a RabbitMQ container """
+    if not ctx.config.get('rabbitmq'):
+        warning('Message queue disabled! This means that the server '
+                'application cannot scale horizontally!')
+    else:
+        # kick off RabbitMQ container
+        rabbit_mgr = RabbitMQManager(
+            ctx=ctx, network_mgr=network_mgr, image=rabbitmq_image)
+        rabbit_mgr.start()
+
+
 #
 #   list
 #
@@ -354,7 +417,7 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
     if image is None:
         image = ctx.config.get(
             "image",
-            "harbor.vantage6.ai/infrastructure/server:latest"
+            f"{DEFAULT_DOCKER_REGISTRY}/{DEFAULT_SERVER_IMAGE}"
         )
     info(f"Pulling latest server image '{image}'.")
     try:
@@ -445,35 +508,32 @@ def cli_server_import(ctx, file_, drop_all, image, keep):
     # fixture.load(entities, drop_all=drop_all)
 
 
-# DISABLED for now - use vserver-local instead
 #
 #   shell
 #
-# @cli_server.command(name='shell')
-# @click_insert_context
-# def cli_server_shell(ctx):
-#     """ Run a iPython shell. """
-#     # make db models available in shell
-#     try:
-#         from vantage6.server import db
-#     except ImportError:
-#         error("vantage6-server not installed")
-#         error(f"install using {Fore.RED}pip install "
-#               f"vantage6-server{Style.RESET_ALL}")
-#         exit(1)
-#
-#     c = get_config()
-#     c.InteractiveShellEmbed.colors = "Linux"
-#
-#     # Suppress logging (e.g. on tab-completion)
-#     import logging
-#     logging.getLogger('parso.cache').setLevel(logging.WARNING)
-#     logging.getLogger('parso.python.diff').setLevel(logging.WARNING)
-#     logging.getLogger('urllib3.connectionpool').setLevel(logging.WARNING)
-#     logging.getLogger('asyncio').setLevel(logging.WARNING)
-#     del logging
-#
-#     IPython.embed(config=c)
+@cli_server.command(name='shell')
+@click_insert_context
+def cli_server_shell(ctx):
+    """ Run a iPython shell. """
+
+    docker_client = docker.from_env()
+    # will print an error if not
+    check_if_docker_deamon_is_running(docker_client)
+
+    running_servers = docker_client.containers.list(
+        filters={"label": f"{APPNAME}-type=server"})
+
+    if not ctx.docker_container_name in [s.name for s in running_servers]:
+        error(f"Server {Fore.RED}{ctx.name}{Style.RESET_ALL} is not running?")
+        return
+
+    try:
+        subprocess.run(['docker', 'exec', '-it', ctx.docker_container_name,
+                        'vserver-local', 'shell', '-c', '/mnt/config.yaml'])
+    except Exception as e:
+        info("Failed to start subprocess...")
+        debug_msg(e)
+
 
 #
 #   stop
@@ -500,26 +560,46 @@ def cli_server_stop(name, system_folders, all_servers):
     running_server_names = [server.name for server in running_servers]
 
     if all_servers:
-        for name in running_server_names:
-            container = client.containers.get(name)
-            container.kill()
-            info(f"Stopped the {Fore.GREEN}{name}{Style.RESET_ALL} server.")
+        for container_name in running_server_names:
+            _stop_server_containers(client, container_name, system_folders)
     else:
         if not name:
-            name = q.select("Select the server you wish to stop:",
-                            choices=running_server_names).ask()
+            container_name = q.select("Select the server you wish to stop:",
+                                      choices=running_server_names).ask()
         else:
 
             post_fix = "system" if system_folders else "user"
-            name = f"{APPNAME}-{name}-{post_fix}-server"
+            container_name = f"{APPNAME}-{name}-{post_fix}-server"
 
-        if name in running_server_names:
-            container = client.containers.get(name)
-            container.kill()
-            info(f"Stopped the {Fore.GREEN}{name}{Style.RESET_ALL} server.")
+        if container_name in running_server_names:
+            _stop_server_containers(client, container_name, system_folders)
         else:
             error(f"{Fore.RED}{name}{Style.RESET_ALL} is not running?")
 
+
+def _stop_server_containers(client: DockerClient, container_name: str,
+                            system_folders: bool) -> None:
+    """
+    Given a server's name, kill its docker container and related (RabbitMQ)
+    containers.
+    """
+    # kill the server
+    container = client.containers.get(container_name)
+    container.kill()
+    info(f"Stopped the {Fore.GREEN}{container_name}{Style.RESET_ALL} server.")
+
+    # find the configuration name from the docker container name
+    # server name is formatted as f"{APPNAME}-{self.name}-{self.scope}-server"
+    scope = "system" if system_folders else "user"
+    config_name = get_server_config_name(container_name, scope)
+
+    # kill the RabbitMQ container (if it exists)
+    rabbit_container_name = f'{APPNAME}-{config_name}-rabbitmq'
+    remove_container_if_exists(client, name=rabbit_container_name)
+    info(f"Stopped the {Fore.GREEN}{rabbit_container_name}{Style.RESET_ALL} "
+         "container.")
+
+
 #
 #   attach
 #
